import json
import logging
import os
import time
from typing import TYPE_CHECKING

import lazy_import
import psutil

from ..config.configs import ClientInfo, ProcessProbeConfig
from ..connector.base_connector import BaseConnector
from ..lang.datamodel_enum import EnvironmentEnum
from ..utils.qoa_utils import (
    convert_to_mbyte,
    get_process_allowed_cpus,
    get_process_allowed_memory,
    report_proc_child_cpu,
    report_proc_mem,
)
from .probe import Probe

if TYPE_CHECKING:
    from ..reports import resources_report_model
else:
    resources_report_model = lazy_import.lazy_module(
        "qoa4ml.reports.resources_report_model"
    )
logging.basicConfig(
    format="%(asctime)s:%(levelname)s -- %(message)s", level=logging.INFO
)


class ProcessMonitoringProbe(Probe):
    def __init__(
        self,
        config: ProcessProbeConfig,
        connector: BaseConnector,
        client_info: ClientInfo,
    ) -> None:
        super().__init__(config, connector, client_info)
        self.config = config
        if self.config.pid is None:
            self.pid = os.getpid()
        else:
            self.pid = self.config.pid
            if not psutil.pid_exists(self.pid):
                raise RuntimeError(f"No process with pid {self.pid}")
        self.environment = config.environment
        self.process = psutil.Process(self.pid)
        if self.config.require_register:
            self.obs_service_url = self.config.obs_service_url
        if self.environment == EnvironmentEnum.hpc:
            self.metadata = {"pid": str(self.pid), "user": self.process.username()}
        else:
            self.metadata = resources_report_model.ProcessMetadata(
                pid=str(self.pid), user=self.process.username(), client_info=client_info
            )

    def get_cpu_usage(self):
        process_usage = report_proc_child_cpu(self.process)
        return process_usage

    def get_mem_usage(self):
        data = report_proc_mem(self.process)
        return {
            "rss": {"value": convert_to_mbyte(data["rss"]), "unit": "Mb"},
            "vms": {"value": convert_to_mbyte(data["vms"]), "unit": "Mb"},
        }

    def create_report(self):
        timestamp = time.time()
        cpu_usage = self.get_cpu_usage()
        mem_usage = self.get_mem_usage()
        allowed_cpu_list = get_process_allowed_cpus()
        allowed_memory_size = get_process_allowed_memory()
        report = None
        if self.environment == EnvironmentEnum.hpc:
            report = {
                "type": "process",
                "metadata": {
                    "pid": str(self.pid),
                    "user": self.process.username(),
                    "allowed_cpu_list": str(allowed_cpu_list),
                    "allowed_memory_size": str(allowed_memory_size),
                },
                "timestamp": round(timestamp),
                "cpu": {
                    "usage": cpu_usage,
                },
                "mem": {
                    "usage": mem_usage,
                },
            }
        else:
            report = resources_report_model.ProcessReport(
                metadata=self.metadata,
                timestamp=round(timestamp),
                cpu=resources_report_model.ResourceReport(usage=cpu_usage),
                mem=resources_report_model.ResourceReport(usage=mem_usage),
            ).model_dump()
        return json.dumps(report)



import copy
import time
from uuid import UUID

from ..config.configs import ClientInfo
from ..lang.common_models import Metric
from ..lang.datamodel_enum import ReportTypeEnum
from ..reports.ml_report_model import (
    FlattenMetric,
    GeneralApplicationReportModel,
    MicroserviceInstance,
)
from .abstract_report import AbstractReport


class GeneralApplicationReport(AbstractReport):
    def __init__(self, client_config: ClientInfo):
        self.client_config = copy.deepcopy(client_config)
        self.reset()
        self.init_time = time.time()

    def reset(self):
        self.report = GeneralApplicationReportModel()

        self.execution_instance = MicroserviceInstance(
            id=UUID(self.client_config.instance_id),
            name=self.client_config.name,
            functionality=self.client_config.functionality,
            stage=self.client_config.stage_id,
        )
        self.previous_reports: list[MicroserviceInstance] = []

    def process_previous_report(self, previous_report_dict: dict):
        previous_report = GeneralApplicationReportModel(**previous_report_dict)
        # NOTE: assume that the last metric is observed by the previous instance
        self.previous_reports.append(previous_report.metrics[-1].instance)
        for metric in previous_report.metrics:
            self.report.metrics.append(metric)

    def observe_metric(self, report_type, stage, metric: Metric):
        flatten_metric = FlattenMetric(
            metric_name=metric.metric_name,
            records=metric.records,
            unit=metric.unit,
            description=metric.description,
            stage=stage,
            report_type=report_type,
            instance=self.execution_instance,
            previous_instances=self.previous_reports,
        )
        self.report.metrics.append(flatten_metric)

    def observe_inference(self, inference_value):
        # TODO: may not be a great idea
        flatten_metric = FlattenMetric(
            metric_name="Inference",
            records=inference_value,
            stage=self.client_config.stage_id,
            report_type=ReportTypeEnum.ml_specific,
            instance=self.execution_instance,
            previous_instances=self.previous_reports,
        )
        self.report.metrics.append(flatten_metric)

    def observe_inference_metric(self, metric: Metric):
        flatten_metric = FlattenMetric(
            metric_name=metric.metric_name,
            records=metric.records,
            unit=metric.unit,
            description=metric.description,
            stage=self.client_config.stage_id,
            report_type=ReportTypeEnum.ml_specific,
            instance=self.execution_instance,
            previous_instances=self.previous_reports,
        )
        self.report.metrics.append(flatten_metric)



import pickle
import socket
from typing import Callable

from ..config.configs import SocketCollectorConfig
from .base_collector import BaseCollector


class SocketCollector(BaseCollector):
    def __init__(self, config: SocketCollectorConfig, process_report: Callable) -> None:
        self.config = config
        self.host = config.host
        self.port = config.port
        self.backlog = config.backlog
        self.bufsize = config.bufsize
        self.process_report = process_report
        self.execution_flag = True

    def start_collecting(self):
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.bind((self.host, self.port))
        server_socket.listen(self.backlog)
        while self.execution_flag:
            client_socket, _ = server_socket.accept()
            data = b""
            while True:
                packet = client_socket.recv(self.bufsize)
                if not packet:
                    break
                data += packet

            report = pickle.loads(data)
            self.process_report(report)
            client_socket.close()



import prometheus_client as pr

_INF = float("inf")


class PromConnector:
    def __init__(self, info):
        self.info = info["metric"]
        self.port = info["port"]
        self.metrices = {}
        for key in self.info:
            self.metrices[key] = {}
            if self.info[key]["Type"] == "Gauge":
                self.metrices[key]["metric"] = pr.Gauge(
                    self.info[key]["Prom_name"], self.info[key]["Description"]
                )
            if self.info[key]["Type"] == "Counter":
                self.metrices[key]["metric"] = pr.Counter(
                    self.info[key]["Prom_name"], self.info[key]["Description"]
                )
            if self.info[key]["Type"] == "Summary":
                self.metrices[key]["metric"] = pr.Summary(
                    self.info[key]["Prom_name"], self.info[key]["Description"]
                )
            if self.info[key]["Type"] == "Histogram":
                self.metrices[key]["metric"] = pr.Histogram(
                    self.info[key]["Prom_name"],
                    self.info[key]["Description"],
                    buckets=(tuple(self.info[key]["Buckets"])),
                )
            self.metrices[key]["violation"] = pr.Counter(
                self.info[key]["Prom_name"] + "_violation",
                self.info[key]["Description"] + " (violation)",
            )
        pr.start_http_server(int(info["port"]))

    def inc(self, key, num=1):
        if self.info[key]["Type"] in ["Gauge", "Counter"]:
            self.metrices[key]["metric"].inc(num)

    def dec(self, key, num=1):
        if self.info[key]["Type"] == "Gauge":
            self.metrices[key]["metric"].dec(num)

    def set(self, key, num=1):
        if self.info[key]["Type"] == "Gauge":
            self.metrices[key]["metric"].set(num)
        elif self.info[key]["Type"] == "Counter":
            self.metrices[key]["metric"].inc(num)
        elif self.info[key]["Type"] in ["Histogram", "Summary"]:
            self.metrices[key]["metric"].observe(num)

    def observe(self, key, val):
        if self.info[key]["Type"] in ["Summary", "Histogram"]:
            self.metrices[key]["metric"].observe(val)

    def inc_violation(self, key, num=1):
        self.metrices[key]["violation"].inc(num)

    def update_violation_count(self):
        for key in self.metrices:
            pr.generate_latest(self.metrices[key]["violation"])



import logging
import socket
from datetime import datetime
from threading import Thread
from typing import TYPE_CHECKING

import lazy_import
from fastapi import APIRouter
from flatten_dict import flatten, unflatten

from ...collector.socket_collector import SocketCollector
from ...config.configs import NodeAggregatorConfig
from ...lang.datamodel_enum import EnvironmentEnum
from ...utils.qoa_utils import make_folder
from .embedded_database import EmbeddedDatabase

logging.basicConfig(
    format="%(asctime)s:%(levelname)s -- %(message)s", level=logging.INFO
)

METRICS_URL_PATH = "/metrics"

if TYPE_CHECKING:
    from ...reports.resources_report_model import ProcessReport, SystemReport
else:
    ProcessReport = lazy_import.lazy_class(
        "...reports.resources_report_model", "ProcessReport"
    )
    SystemReport = lazy_import.lazy_class(
        "...reports.resources_report_model", "SystemReport"
    )


class NodeAggregator:
    def __init__(self, config: NodeAggregatorConfig):
        self.config = config
        self.unit_conversion = self.config.unit_conversion
        self.node_name = socket.gethostname().split(".")[0]
        make_folder(self.config.database_path)
        self.embedded_database = EmbeddedDatabase(
            self.config.database_path + self.node_name + ".csv"
        )
        self.environment = config.environment
        self.collector = SocketCollector(
            config.socket_collector_config, self.process_report
        )
        self.server_thread = Thread(target=self.collector.start_collecting, daemon=True)
        self.router = APIRouter()
        self.router.add_api_route(
            METRICS_URL_PATH,
            self.get_lastest_timestamp,
            methods=[self.config.query_method],
        )

    def process_report(self, report):
        if self.environment == EnvironmentEnum.hpc:
            if report["type"] == "system":
                del report["type"]
                metadata = flatten(
                    {"metadata": report["metadata"]}, self.config.data_separator
                )
                timestamp = report["timestamp"]
                del report["metadata"], report["timestamp"]
                fields = self.convert_unit(flatten(report, self.config.data_separator))
                self.embedded_database.insert(
                    timestamp,
                    {"type": "node", **metadata},
                    fields,
                )
            elif report["type"] == "process":
                del report["type"]
                metadata = flatten(
                    {"metadata": report["metadata"]}, self.config.data_separator
                )
                timestamp = report["timestamp"]
                del report["metadata"], report["timestamp"]
                fields = self.convert_unit(flatten(report, self.config.data_separator))
                self.embedded_database.insert(
                    timestamp,
                    {"type": "process", **metadata},
                    fields,
                )
            else:
                logging.error("Value Error: Unknown report type")
        elif isinstance(report, SystemReport):
            node_name = report.metadata.node_name
            timestamp = report.timestamp
            del report.metadata, report.timestamp
            fields = self.convert_unit(
                flatten(report.dict(exclude_none=True), self.config.data_separator)
            )
            self.embedded_database.insert(
                timestamp,
                {
                    "type": "node",
                    "node_name": node_name,
                },
                fields,
            )
        elif isinstance(report, ProcessReport):
            metadata = flatten(
                {"metadata": report.metadata.dict()}, self.config.data_separator
            )
            timestamp = report.timestamp
            del report.metadata, report.timestamp
            fields = self.convert_unit(
                flatten(report.dict(exclude_none=True), self.config.data_separator)
            )
            self.embedded_database.insert(
                timestamp, {"type": "process", **metadata}, fields
            )

    def convert_unit(self, report: dict):
        converted_report = report
        for key, value in report.items():
            # TODO: for instead of hard coded
            if isinstance(value, str):
                if "frequency" in key:
                    converted_report[key] = self.unit_conversion["frequency"][value]
                elif "mem" in key:
                    converted_report[key] = self.unit_conversion["mem"][value]
                elif "cpu" in key:
                    if "usage" in key:
                        converted_report[key] = self.unit_conversion["cpu"]["usage"][
                            value
                        ]
                elif "gpu" in key:
                    if "usage" in key:
                        converted_report[key] = self.unit_conversion["gpu"]["usage"][
                            value
                        ]
        return converted_report

    def revert_unit(self, converted_report: dict):
        original_report = converted_report.copy()
        for key, value in converted_report.items():
            if "unit" in key:
                if "frequency" in key:
                    for original_unit, converted_unit in self.unit_conversion[
                        "frequency"
                    ].items():
                        if converted_unit == value:
                            original_report[key] = original_unit
                            break
                elif "mem" in key:
                    for original_unit, converted_unit in self.unit_conversion[
                        "mem"
                    ].items():
                        if converted_unit == value:
                            original_report[key] = original_unit
                            break
                elif "cpu" in key:
                    if "usage" in key:
                        for original_unit, converted_unit in self.unit_conversion[
                            "cpu"
                        ]["usage"].items():
                            if converted_unit == value:
                                original_report[key] = original_unit
                                break
                elif "gpu" in key:
                    if "usage" in key:
                        for original_unit, converted_unit in self.unit_conversion[
                            "gpu"
                        ]["usage"].items():
                            if converted_unit == value:
                                original_report[key] = original_unit
                                break
        return original_report

    def get_lastest_timestamp(self):
        data = self.embedded_database.get_lastest_timestamp()
        return [
            unflatten(
                self.revert_unit(
                    {
                        "timestamp": datetime.timestamp(datapoint.time),
                        **datapoint.tags,
                        **datapoint.fields,
                    }
                ),
                self.config.data_separator,
            )
            for datapoint in data
        ]

    def start(self):
        self.execution_flag = True
        self.server_thread.start()
        logging.info("node aggregator started")

    def stop(self):
        self.execution_flag = False
        self.server_thread.join()
        logging.info("node aggregator stopped")



import copy
import os
import sys
import threading
import time
import traceback
import uuid
from threading import Thread
from typing import Any, Generic, Optional, TypeVar, Union

import requests
from pydantic import create_model

# from .connector.mqtt_connector import Mqtt_Connector
from .config.configs import (
    AMQPConnectorConfig,
    ClientConfig,
    ClientInfo,
    ConnectorConfig,
    DebugConnectorConfig,
    DockerProbeConfig,
    ProbeConfig,
    ProcessProbeConfig,
    SystemProbeConfig,
)
from .connector.amqp_connector import AmqpConnector
from .connector.base_connector import BaseConnector
from .connector.debug_connector import DebugConnector
from .lang.common_models import Metric
from .lang.datamodel_enum import (
    MetricNameEnum,
    ReportTypeEnum,
    ServiceAPIEnum,
    ServiceMetricNameEnum,
)
from .probes.docker_monitoring_probe import DockerMonitoringProbe
from .probes.probe import Probe
from .probes.process_monitoring_probe import ProcessMonitoringProbe
from .probes.system_monitoring_probe import SystemMonitoringProbe
from .reports.abstract_report import AbstractReport
from .reports.ml_reports import MLReport
from .utils.logger import qoa_logger
from .utils.qoa_utils import (
    load_config,
    set_logger_level,
)

headers = {"Content-Type": "application/json"}


# NOTE: T must be a subtype of AbstractReport
T = TypeVar("T", bound=AbstractReport)


class QoaClient(Generic[T]):
    def __init__(
        self,
        # NOTE: use text, number, enum
        report_cls: type[T] = MLReport,
        config_dict: Optional[dict] = None,
        config_path: Optional[str] = None,
        registration_url: Optional[str] = None,
        logging_level=2,
    ):
        set_logger_level(logging_level)

        if config_dict is not None:
            self.configuration = ClientConfig.model_validate(config_dict)

        if config_path is not None:
            self.configuration = ClientConfig.model_validate(load_config(config_path))

        self.client_config = self.configuration.client
        self.connector_list: dict[str, BaseConnector] = {}
        self.timer_flag = False
        self.functionality = self.client_config.functionality
        self.stage_id = self.client_config.stage_id
        self.process_monitor_flag = 0
        self.inference_flag = False

        self.instance_id = os.environ.get("INSTANCE_ID")
        if self.instance_id:
            qoa_logger.info("Setting instance_id with INSTANCE_ID")
            self.instance_id = uuid.UUID(self.instance_id)
        elif self.configuration.client.instance_id:
            qoa_logger.info("Setting instance_id with client.instance_id")
            self.instance_id = uuid.UUID(self.configuration.client.instance_id)
        else:
            self.instance_id = str(uuid.uuid4())

        self.client_config.instance_id = str(self.instance_id)
        self.client_config.id = str(uuid.uuid4())
        self.qoa_report = report_cls(self.client_config)
        if self.configuration.connector:
            # init connectors offline if it's specified
            connector_conf = self.configuration.connector
            try:
                for connector in connector_conf:
                    self.connector_list[connector.name] = self.init_connector(connector)
            except Exception as e:
                qoa_logger.exception(
                    f"Error {type(e)} when configuring connector in QoaClient"
                )
        elif registration_url or self.configuration.registration_url:
            # init connectors using configuration received from monitoring service, if it's specified
            try:
                if registration_url:
                    registration_data = self.registration(registration_url)
                else:
                    # NOTE: logically true
                    registration_data = self.registration(
                        self.configuration.registration_url
                    )
                json_data = registration_data.json()
                response = json_data["response"]
                if isinstance(response, dict):
                    connector_configs = response["connector"]
                    if isinstance(connector_configs, dict):
                        connector_config = ConnectorConfig(**connector_configs)
                        self.connector_list[connector_config.name] = (
                            self.init_connector(connector_config)
                        )
                    elif isinstance(connector_configs, list):
                        for config in connector_configs:
                            connector_config = ConnectorConfig(**config)
                            self.connector_list[connector_config.name] = (
                                self.init_connector(connector_config)
                            )
                else:
                    qoa_logger.warning(
                        "Unable to register Qoa Client: connector configuration must be dictionary"
                    )
            except Exception as e:
                qoa_logger.exception(f"Error {type(e)} when registering QoA client")
                traceback.print_exception(*sys.exc_info())

        if not self.connector_list:
            qoa_logger.warning("No connector initiated")
            self.default_connector = None
        else:
            # Set default connector for sending monitoring data if not specify
            self.default_connector = next(iter(self.connector_list.keys()))

        self.probes_list = None
        if self.configuration.probes:
            self.probes_list = self.init_probes(
                self.configuration.probes, self.configuration.client
            )
        # lock report to guarantee consistency
        self.lock = threading.Lock()

    def registration(self, url: str):
        # get connector configuration by registering with the monitoring service

        return requests.request(
            "POST", url, headers=headers, data=self.client_config.json()
        )

    def init_probes(
        self, probe_config_list: list[ProbeConfig], client_info: ClientInfo
    ):
        probes_list: list[Probe] = []
        # TODO: each probe can have their own connector
        if self.default_connector:
            selected_connector = self.connector_list[self.default_connector]
        else:
            qoa_logger.warning("No default connector, using debug connector")
            selected_connector = DebugConnector()

        for probe_config in probe_config_list:
            # TODO: can be simplify for less duplicate code
            if isinstance(probe_config, DockerProbeConfig):
                probes_list.append(
                    DockerMonitoringProbe(probe_config, selected_connector, client_info)
                )
            elif isinstance(probe_config, ProcessProbeConfig):
                probes_list.append(
                    ProcessMonitoringProbe(
                        probe_config, selected_connector, client_info
                    )
                )
            elif isinstance(probe_config, SystemProbeConfig):
                probes_list.append(
                    SystemMonitoringProbe(probe_config, selected_connector, client_info)
                )
            else:
                raise ValueError(
                    f"Probe config type {type(probe_config)} is not supported yet"
                )
        return probes_list

    def init_connector(self, configuration: ConnectorConfig) -> BaseConnector:
        # init connector from configuration
        if configuration.connector_class == ServiceAPIEnum.amqp and isinstance(
            configuration.config, AMQPConnectorConfig
        ):
            return AmqpConnector(configuration.config)
        elif configuration.connector_class == ServiceAPIEnum.debug and isinstance(
            configuration.config, DebugConnectorConfig
        ):
            return DebugConnector(configuration.config)

        # TODO: MQTT is both connector and collector
        #
        # if (
        #    configuration.connector_class == ServiceAPIEnum.mqtt
        #    and type(configuration.config) is MQTTConnectorConfig
        # ):
        #    return Mqtt_Connector(configuration.config)
        raise RuntimeError("Connector config is not of correct type")

    def get_client_config(self):
        # TODO: what to do exactly?
        return self.client_config

    def set_config(self, key, value):
        try:
            self.client_config.__setattr__(key, value)
        except Exception as e:
            qoa_logger.exception(f"Error {type(e)} when setConfig in QoA client")

    def observe_metric(
        self,
        metric_name: MetricNameEnum,
        value,
        category: int = 0,
        description: str = "",
    ):
        if category == 0:
            report_type = ReportTypeEnum.service
        elif category == 1:
            report_type = ReportTypeEnum.data
        elif category == 2:
            report_type = ReportTypeEnum.security
        else:
            raise RuntimeError("Report type not supported")

        self.qoa_report.observe_metric(
            report_type,
            self.stage_id,
            Metric(metric_name=metric_name, records=[value], description=description),
        )

    def timer(self):
        if self.timer_flag is False:
            self.timer_flag = True
            self.timerStart = time.time()
            return {}
        else:
            self.timer_flag = False
            response_time = {
                "startTime": self.timerStart,
                "responseTime": time.time() - self.timerStart,
            }
            self.observe_metric(
                ServiceMetricNameEnum.response_time, response_time, category=0
            )
            return response_time

    def import_previous_report(self, reports: Union[dict, list[dict]]):
        if isinstance(reports, list):
            for report in reports:
                self.qoa_report.process_previous_report(report)
        else:
            self.qoa_report.process_previous_report(reports)

    def __str__(self):
        return self.client_config.model_dump_json() + "\n" + str(self.connector_list)

    def asyn_report(self, body_mess: str, connectors: Optional[list] = None):
        self.lock.acquire()
        if connectors is None:
            # if connectors are not specify, use default
            if self.default_connector:
                chosen_connector = self.connector_list[self.default_connector]
                if isinstance(chosen_connector, AmqpConnector):
                    chosen_connector.send_report(body_mess, corr_id=str(uuid.uuid4()))
                else:
                    chosen_connector.send_report(body_mess)
            else:
                qoa_logger.error(
                    "No default connector, please specify the connector to use"
                )
        else:
            # iterate connector to send report
            # for connector in connectors:
            # print(connector)
            # Todo: send by multiple connector
            pass

        self.lock.release()

    def report(
        self,
        report: Optional[dict] = None,
        connectors: Optional[list] = None,
        submit=False,
        reset=True,
        corr_id=None,
    ):
        """
        submit=True to submit the report through the default connector
        """
        if report is None:
            return_report = self.qoa_report.generate_report(reset, corr_id=corr_id)
        else:
            user_defined_report_model = create_model(
                "UserDefinedReportModel",
                metadata=(dict, ...),
                timestamp=(float, ...),
                report=(dict, ...),
            )
            return_report = user_defined_report_model(
                report=report,
                metadata=copy.deepcopy(self.client_config.__dict__),
                timestamp=time.time(),
            )

        if submit:
            if self.default_connector is not None:
                sub_thread = Thread(
                    target=self.asyn_report,
                    args=(return_report.model_dump_json(), connectors),
                )
                sub_thread.start()
            else:
                qoa_logger.warning("No connector available")
        return return_report.model_dump(mode="json")

    def start_all_probes(self):
        """
        Start all probes in the background, will be killed when the main process exited
        NOTE: if the probe takes long to report, and the main process exit, no report may be sent
        """
        if not self.probes_list:
            raise RuntimeError(
                "There is no initiated probes, please recheck the config"
            )
        for probe in self.probes_list:
            probe.start_reporting()

    def stop_all_probes(self):
        if not self.probes_list:
            raise RuntimeError(
                "There is no initiated probes, please recheck the config"
            )
        for probe in self.probes_list:
            probe.stop_reporting()

    def observe_inference(
        self,
        inference_value,
    ):
        self.qoa_report.observe_inference(inference_value)

    def observe_inference_metric(
        self,
        metric_name: MetricNameEnum,
        value: Any,
    ):
        metric = Metric(metric_name=metric_name, records=[value])
        self.qoa_report.observe_inference_metric(metric)



import json
from typing import Optional

import pika

from ..config.configs import AMQPCollectorConfig
from ..utils.logger import qoa_logger
from .base_collector import BaseCollector
from .host_object import HostObject


class AmqpCollector(BaseCollector):
    # Init an amqp client handling the connection to amqp servier
    def __init__(
        self,
        configuration: AMQPCollectorConfig,
        host_object: Optional[HostObject] = None,
    ):
        self.host_object = host_object
        self.exchange_name = configuration.exchange_name
        self.exchange_type = configuration.exchange_type
        self.in_routing_key = configuration.in_routing_key

        # Connect to RabbitMQ host
        if "amqps://" in configuration.end_point:
            self.in_connection = pika.BlockingConnection(
                pika.URLParameters(configuration.end_point)
            )
        else:
            self.in_connection = pika.BlockingConnection(
                pika.ConnectionParameters(host=configuration.end_point)
            )

        # Create a channel
        self.in_channel = self.in_connection.channel()

        # Init an Exchange
        self.in_channel.exchange_declare(
            exchange=self.exchange_name, exchange_type=self.exchange_type
        )

        # Declare a queue to receive prediction response
        self.queue = self.in_channel.queue_declare(
            queue=configuration.in_queue, exclusive=False
        )
        self.queue_name = self.queue.method.queue
        # Binding the exchange to the queue with specific routing
        self.in_channel.queue_bind(
            exchange=self.exchange_name,
            queue=self.queue_name,
            routing_key=self.in_routing_key,
        )

    def on_request(self, ch, method, props, body):
        # Process the data on request: sending back to host object
        if self.host_object is not None:
            self.host_object.message_processing(ch, method, props, body)
        else:
            mess = json.loads(str(body.decode("utf-8")))
            qoa_logger.info(mess)

    def start_collecting(self):
        # Start rabbit MQ
        self.in_channel.basic_qos(prefetch_count=1)
        self.in_channel.basic_consume(
            queue=self.queue_name, on_message_callback=self.on_request, auto_ack=True
        )
        self.in_channel.start_consuming()

    def stop(self):
        self.in_channel.stop_consuming()
        self.in_channel.close()

    def get_queue(self):
        return self.queue.method.queue



import copy
import time
from typing import Optional
from uuid import UUID, uuid4

from ..config.configs import ClientInfo
from ..lang.common_models import Metric
from ..lang.datamodel_enum import ReportTypeEnum
from ..reports.abstract_report import AbstractReport
from ..reports.ml_report_model import (
    EnsembleInferenceReport,
    ExecutionGraph,
    InferenceGraph,
    InferenceInstance,
    LinkedInstance,
    MicroserviceInstance,
    RoheReportModel,
    StageReport,
)
from ..utils.qoa_utils import load_config


class RoheReport(AbstractReport):
    def __init__(self, client_config: ClientInfo):
        self.client_config = copy.deepcopy(client_config)
        self.reset()
        self.init_time = time.time()

        # if file_path:
        #     self.import_report_from_file(file_path)

    def reset(self):
        self.previous_report: list[RoheReportModel] = []
        self.inference_report = EnsembleInferenceReport()
        self.execution_graph = ExecutionGraph(linked_list={})
        self.report = RoheReportModel()
        self.previous_microservice_instance = []
        self.execution_instance = MicroserviceInstance(
            id=UUID(self.client_config.instance_id),
            name=self.client_config.name,
            functionality=self.client_config.functionality,
            stage=self.client_config.stage_id,
        )

    def import_report_from_file(self, file_path: str):
        report = load_config(file_path)
        self.inference_report = EnsembleInferenceReport(**report["inference_report"])
        self.execution_graph = ExecutionGraph(**report["execution_graph"])
        self.report = RoheReportModel(
            inference_report=self.inference_report, execution_graph=self.execution_graph
        )

    def combine_stage_report(
        self,
        current_stage_report: dict[str, StageReport],
        previous_stage_report: dict[str, StageReport],
    ):
        combined_stage_report: dict[str, StageReport] = {}
        for stage_name, stage_report in previous_stage_report.items():
            new_stage_report = StageReport(name=stage_name, metrics={})
            if stage_name not in current_stage_report:
                current_stage_report[stage_name] = StageReport(
                    name=stage_name, metrics={}
                )
            for metric_name, instance_report_dict in stage_report.metrics.items():
                if metric_name not in current_stage_report[stage_name].metrics:
                    current_stage_report[stage_name].metrics[metric_name] = {}
                new_stage_report.metrics[metric_name] = {
                    **current_stage_report[stage_name].metrics[metric_name],
                    **instance_report_dict,
                }
            combined_stage_report[stage_name] = new_stage_report
        return combined_stage_report

    def process_previous_report(self, previous_report_dict: dict):
        previous_report = RoheReportModel(**previous_report_dict)
        self.previous_report.append(previous_report)
        if not previous_report.inference_report or not previous_report.execution_graph:
            raise ValueError("Can't process empty previous report")

        # NOTE: service quality report
        self.inference_report.service = self.combine_stage_report(
            self.inference_report.service, previous_report.inference_report.service
        )

        # NOTE: data quality report
        self.inference_report.data = self.combine_stage_report(
            self.inference_report.data, previous_report.inference_report.data
        )

        # NOTE: ml-specific quality report
        # debug(previous_report.inference_report.ml_specific)
        if not self.inference_report.ml_specific:
            if (
                previous_report.inference_report.ml_specific
                and previous_report.inference_report.ml_specific.end_point
                and previous_report.inference_report.ml_specific.linked_list
            ):
                self.inference_report.ml_specific = (
                    previous_report.inference_report.ml_specific
                )
                end_point = InferenceInstance(
                    inference_id=uuid4(),
                    functionality=self.client_config.functionality,
                    instance_id=self.execution_instance.id,
                )
                self.inference_report.ml_specific.end_point = end_point
                self.inference_report.ml_specific.linked_list |= {
                    end_point.instance_id: LinkedInstance[InferenceInstance](
                        instance=end_point,
                        previous=[
                            previous_report.inference_report.ml_specific.end_point
                        ],
                    )
                }

        elif (
            previous_report.inference_report.ml_specific
            and previous_report.inference_report.ml_specific.end_point
            and previous_report.inference_report.ml_specific.linked_list
            and self.inference_report.ml_specific.end_point
        ):
            self.inference_report.ml_specific.linked_list |= (
                previous_report.inference_report.ml_specific.linked_list
            )
            current_end_point = self.inference_report.ml_specific.end_point
            previous_end_point = previous_report.inference_report.ml_specific.end_point
            self.inference_report.ml_specific.linked_list[
                current_end_point.instance_id
            ].previous.append(previous_end_point)
        # NOTE: execution graph
        if not self.execution_graph:
            self.execution_graph = previous_report.execution_graph
        else:
            self.execution_graph.linked_list.update(
                previous_report.execution_graph.linked_list
            )
        self.previous_microservice_instance.append(
            previous_report.execution_graph.end_point
        )
        self.report = RoheReportModel(
            inference_report=self.inference_report, execution_graph=self.execution_graph
        )

    def build_execution_graph(self):
        end_point = LinkedInstance(
            instance=self.execution_instance,
            previous=list(self.previous_microservice_instance),
        )
        self.execution_graph.linked_list[end_point.instance.id] = end_point
        self.execution_graph.end_point = end_point.instance
        self.report.execution_graph = self.execution_graph

    def observe_metric(self, report_type: ReportTypeEnum, stage: str, metric: Metric):
        if stage == "":
            raise ValueError("Stage name can't be empty")
        if report_type == ReportTypeEnum.service:
            if stage not in self.inference_report.service:
                self.inference_report.service[stage] = StageReport(
                    name=stage, metrics={}
                )
            if metric.metric_name not in self.inference_report.service[stage].metrics:
                self.inference_report.service[stage].metrics[metric.metric_name] = {}

            self.inference_report.service[stage].metrics[metric.metric_name] |= {
                UUID(self.client_config.instance_id): metric
            }

        elif report_type == ReportTypeEnum.data:
            if stage not in self.inference_report.data:
                self.inference_report.data[stage] = StageReport(name=stage, metrics={})
            if metric.metric_name not in self.inference_report.data[stage].metrics:
                self.inference_report.data[stage].metrics[metric.metric_name] = {}

            self.inference_report.data[stage].metrics[metric.metric_name] |= {
                UUID(self.client_config.instance_id): metric
            }
        else:
            raise ValueError(f"Can't handle report type {report_type}")
        self.report.inference_report = self.inference_report

    def observe_inference(self, inference_value):
        if (
            self.inference_report.ml_specific
            and self.inference_report.ml_specific.end_point
        ):
            self.inference_report.ml_specific.end_point.prediction = inference_value
        else:
            self.inference_report.ml_specific = InferenceGraph()
            end_point = InferenceInstance(
                inference_id=uuid4(),
                instance_id=self.execution_instance.id,
                functionality=self.client_config.functionality,
                prediction=inference_value,
            )
            self.inference_report.ml_specific.end_point = end_point
            self.inference_report.ml_specific.linked_list |= {
                end_point.instance_id: LinkedInstance[InferenceInstance](
                    instance=end_point,
                    previous=[],
                )
            }

    def observe_inference_metric(
        self,
        metric: Metric,
    ):
        if (
            self.inference_report.ml_specific
            and self.inference_report.ml_specific.end_point
        ):
            self.inference_report.ml_specific.end_point.metrics.append(metric)
        else:
            if self.inference_report.ml_specific is None:
                self.inference_report.ml_specific = InferenceGraph()
            if self.inference_report.ml_specific.end_point is None:
                end_point = InferenceInstance(
                    inference_id=uuid4(),
                    instance_id=self.execution_instance.id,
                    functionality=self.client_config.functionality,
                )

                self.inference_report.ml_specific.end_point = end_point
                self.inference_report.ml_specific.linked_list[end_point.instance_id] = (
                    LinkedInstance[InferenceInstance](instance=end_point)
                )
            self.inference_report.ml_specific.end_point.metrics.append(metric)

    def generate_report(self, reset: bool = True, corr_id: Optional[str] = None):
        self.build_execution_graph()
        self.report.metadata["client_config"] = copy.deepcopy(self.client_config)
        self.report.metadata["timestamp"] = time.time()
        if corr_id is not None:
            self.report.metadata["corr_id"] = corr_id
        self.report.metadata["runtime"] = (
            self.report.metadata["timestamp"] - self.init_time
        )
        report = copy.deepcopy(self.report)
        if reset:
            self.reset()
        return report



import json

from devtools import debug

from ..config.configs import DebugConnectorConfig
from .base_connector import BaseConnector


class DebugConnector(BaseConnector):
    def __init__(self, config: DebugConnectorConfig):
        self.silence = config.silence

    def send_report(self, body_message: str):
        if not self.silence:
            debug(json.loads(body_message))



import uuid
from typing import Optional

import pika

from ..config.configs import AMQPConnectorConfig
from .base_connector import BaseConnector


class AmqpConnector(BaseConnector):
    # Init an amqp client handling the connection to amqp servier
    def __init__(self, config: AMQPConnectorConfig, log: bool = False):
        """
        AMQP connector
        configuration: a dictionary include broker and queue information
        log: a bool flag for logging message if being set to True, default is False
        """
        self.conf = config
        self.exchange_name = config.exchange_name
        self.exchange_type = config.exchange_type
        self.out_routing_key = config.out_routing_key
        self.log_flag = log

        # Connect to RabbitMQ host
        if "amqps://" in config.end_point:
            self.out_connection = pika.BlockingConnection(
                pika.URLParameters(config.end_point)
            )
        else:
            self.out_connection = pika.BlockingConnection(
                pika.ConnectionParameters(host=config.end_point)
            )

        # Create a channel
        self.out_channel = self.out_connection.channel()

        # Init an Exchange
        self.out_channel.exchange_declare(
            exchange=self.exchange_name, exchange_type=self.exchange_type
        )

    def send_report(
        self,
        body_message: str,
        corr_id=None,
        routing_key: Optional[str] = None,
        expiration=1000,
    ):
        # Sending data to desired destination
        # if sender is client, it will include the "reply_to" attribute to specify where to reply this message
        # if sender is server, it will reply the message to "reply_to" via default exchange
        if corr_id is None:
            corr_id = str(uuid.uuid4())
        if routing_key is None:
            routing_key = self.out_routing_key
        self.sub_properties = pika.BasicProperties(
            correlation_id=corr_id, expiration=str(expiration)
        )
        self.out_channel.basic_publish(
            exchange=self.exchange_name,
            routing_key=routing_key,
            properties=self.sub_properties,
            body=body_message,
        )
        # if self.log_flag:
        #     self.mess_logging.log_request(body_mess,corr_id)

    def get(self):
        return self.conf



import copy
import time
from typing import Optional
from uuid import UUID, uuid4

from ..config.configs import ClientInfo
from ..lang.common_models import Metric
from ..lang.datamodel_enum import ReportTypeEnum
from ..reports.ml_report_model import (
    BaseReport,
    GeneralMlInferenceReport,
    InferenceInstance,
    StageReport,
)
from .abstract_report import AbstractReport


class MLReport(AbstractReport):
    def __init__(self, client_config: ClientInfo):
        self.client_config = copy.deepcopy(client_config)
        self.reset()
        self.init_time = time.time()

    def reset(self):
        self.previous_report: list[GeneralMlInferenceReport] = []
        self.report = GeneralMlInferenceReport()

    def combine_stage_report(
        self,
        current_stage_report: dict[str, StageReport],
        previous_stage_report: dict[str, StageReport],
    ):
        combined_stage_report: dict[str, StageReport] = {}
        for stage_name, stage_report in previous_stage_report.items():
            new_stage_report = StageReport(name=stage_name, metrics={})
            if stage_name not in current_stage_report:
                current_stage_report[stage_name] = StageReport(
                    name=stage_name, metrics={}
                )
            for metric_name, instance_report_dict in stage_report.metrics.items():
                if metric_name not in current_stage_report[stage_name].metrics:
                    current_stage_report[stage_name].metrics[metric_name] = {}
                new_stage_report.metrics[metric_name] = {
                    **current_stage_report[stage_name].metrics[metric_name],
                    **instance_report_dict,
                }
            combined_stage_report[stage_name] = new_stage_report
        return combined_stage_report

    def process_previous_report(self, previous_report_dict: dict):
        previous_report = GeneralMlInferenceReport(**previous_report_dict)
        self.previous_report.append(previous_report)

        # NOTE: service quality report
        self.report.service = self.combine_stage_report(
            self.report.service, previous_report.service
        )

        # NOTE: data quality report
        self.report.data = self.combine_stage_report(
            self.report.data, previous_report.data
        )
        # NOTE: ml inference report
        #
        self.report.ml_inference |= previous_report.ml_inference

    def observe_metric(self, report_type, stage, metric: Metric):
        if stage == "":
            raise ValueError("Stage name can't be empty")
        if report_type == ReportTypeEnum.service:
            if stage not in self.report.service:
                self.report.service[stage] = StageReport(name=stage, metrics={})
            if metric.metric_name not in self.report.service[stage].metrics:
                self.report.service[stage].metrics[metric.metric_name] = {}

            self.report.service[stage].metrics[metric.metric_name] |= {
                UUID(self.client_config.instance_id): metric
            }

        elif report_type == ReportTypeEnum.data:
            if stage not in self.report.data:
                self.report.data[stage] = StageReport(name=stage, metrics={})
            if metric.metric_name not in self.report.data[stage].metrics:
                self.report.data[stage].metrics[metric.metric_name] = {}

            self.report.data[stage].metrics[metric.metric_name] |= {
                UUID(self.client_config.instance_id): metric
            }
        else:
            raise ValueError(f"Can't handle report type {report_type}")

    def observe_inference(self, inference_value):
        if self.client_config.instance_id in self.report.ml_inference:
            raise RuntimeWarning(
                "Inference existed, will override the existing inference"
            )
        self.report.ml_inference[self.client_config.instance_id] = InferenceInstance(
            inference_id=uuid4(),
            instance_id=UUID(self.client_config.instance_id),
            functionality=self.client_config.functionality,
            prediction=inference_value,
        )

    def observe_inference_metric(self, metric: Metric):
        if self.client_config.instance_id in self.report.ml_inference:
            self.report.ml_inference[self.client_config.instance_id].metrics.append(
                metric
            )
        else:
            self.report.ml_inference[self.client_config.instance_id] = (
                InferenceInstance(
                    inference_id=uuid4(),
                    instance_id=UUID(self.client_config.instance_id),
                    functionality=self.client_config.functionality,
                    metrics=[metric],
                )
            )

    def generate_report(
        self, reset: bool = True, corr_id: Optional[str] = None
    ) -> BaseReport:
        self.report.metadata["client_config"] = copy.deepcopy(self.client_config)
        self.report.metadata["timestamp"] = time.time()
        if corr_id is not None:
            self.report.metadata["corr_id"] = corr_id
        self.report.metadata["runtime"] = (
            self.report.metadata["timestamp"] - self.init_time
        )
        report = copy.deepcopy(self.report)
        if reset:
            self.reset()
        return report



from __future__ import annotations

from typing import Union

from pydantic import BaseModel, Field, model_validator

from ..lang.datamodel_enum import (
    EnvironmentEnum,
    MetricClassEnum,
    MetricNameEnum,
    ServiceAPIEnum,
)


class ClientInfo(BaseModel):
    id: str = Field(
        default="",
        description="The id of the QoAClient, should be not be set manually, auto-generated by QoAClient",
    )
    name: str = Field(
        default="",
        description="The name of the QoAClient, for example, qoa_client1",
    )

    user_id: str = Field(
        default="",
        description="Can be set if you have external user_id, otherwise auto generated",
    )
    username: str = Field(default="", description="The name of the qoa_client user")
    instance_id: str = Field(
        default="",
        description="An UUID4 that is used to identify the instance, can be set manually in the config or through environment variable INSTANCE_ID, if not, auto-generated by QoAClient",
    )
    instance_name: str = Field(default="", description="The name of the instance")
    stage_id: str = ""
    functionality: str = ""
    application_name: str = ""
    role: str = ""
    run_id: str = ""
    environment: EnvironmentEnum = Field(
        default=EnvironmentEnum.edge,
        description="The environment where the qoa_client is run",
    )
    custom_info: dict | str = Field(
        default="", description="Additional information that you can add"
    )


class AMQPCollectorConfig(BaseModel):
    end_point: str
    exchange_name: str
    exchange_type: str
    in_routing_key: str
    in_queue: str


class AMQPConnectorConfig(BaseModel):
    end_point: str
    exchange_name: str
    exchange_type: str
    out_routing_key: str


class MQTTConnectorConfig(BaseModel):
    in_queue: str
    out_queue: str
    broker_url: str
    broker_port: int
    broker_keepalive: int
    client_id: str


class SocketConnectorConfig(BaseModel):
    host: str
    port: int


class SocketCollectorConfig(BaseModel):
    host: str
    port: int
    backlog: int
    bufsize: int


class PrometheusConnectorConfig(BaseModel):
    pass


class KafkaConnectorConfig(BaseModel):
    topic: str
    broker_url: str


class KafkaCollectorConfig(BaseModel):
    topic: str
    broker_url: str
    group_id: str
    auto_offset_reset: str = "earliest"
    poll_inteval: float = 1.0


class DebugConnectorConfig(BaseModel):
    silence: bool


# TODO: test if loading the config, the type of the config can be found
CollectorConfigClass = Union[
    AMQPCollectorConfig, SocketCollectorConfig, KafkaCollectorConfig, dict
]
ConnectorConfigClass = Union[
    AMQPConnectorConfig,
    SocketConnectorConfig,
    KafkaConnectorConfig,
    DebugConnectorConfig,
    dict,
]


class CollectorConfig(BaseModel):
    name: str
    collector_class: ServiceAPIEnum
    config: CollectorConfigClass


class ConnectorConfig(BaseModel):
    name: str
    connector_class: ServiceAPIEnum
    config: ConnectorConfigClass | None = None


class ClientConfig(BaseModel):
    client: ClientInfo
    registration_url: str | None = None
    collector: list[CollectorConfig] | None = None
    connector: list[ConnectorConfig] | None = None
    probes: list[ProbeConfig] | None = None

    @model_validator(mode="before")
    @classmethod
    def validate_probe_type(cls, values):
        probe_type_map: dict[str, type[ProbeConfig]] = {
            "process": ProcessProbeConfig,
            "docker": DockerProbeConfig,
            "system": SystemProbeConfig,
        }
        probes = values.get("probes", [])
        if probes:
            for i, probe_data in enumerate(probes):
                probe_type = probe_data.get("probe_type")
                if probe_type in probe_type_map:
                    probes[i] = probe_type_map[probe_type](**probe_data)
            values["probes"] = probes
        return values


class MetricConfig(BaseModel):
    name: MetricNameEnum
    metric_class: MetricClassEnum
    description: str | None = None
    default_value: int
    category: int


class GroupMetricConfig(BaseModel):
    name: str
    metric_configs: list[MetricConfig]


class ProbeConfig(BaseModel):
    probe_type: str
    frequency: int
    require_register: bool
    obs_service_url: str | None = None
    log_latency_flag: bool
    latency_logging_path: str | None = None
    environment: EnvironmentEnum = Field(
        default=EnvironmentEnum.edge, description="The environment where the probe run"
    )


class ProcessProbeConfig(ProbeConfig):
    probe_type: str = Field("process")
    pid: int | None = None


class SystemProbeConfig(ProbeConfig):
    probe_type: str = Field("system")
    node_name: str | None = None


class DockerProbeConfig(ProbeConfig):
    probe_type: str = Field("docker")
    container_list: list[str] = []


class JetsonSystemProbeConfig(ProbeConfig):
    probe_type: str = Field("jetson_sys")
    node_name: str | None


class JetsonProcessesProbeConfig(ProbeConfig):
    probe_type: str = Field("jetson_sys")
    node_name: str | None


class NodeAggregatorConfig(BaseModel):
    socket_collector_config: SocketCollectorConfig
    environment: EnvironmentEnum
    query_method: str
    data_separator: str
    unit_conversion: dict
    database_path: str


class ExporterConfig(BaseModel):
    host: str
    port: int
    node_aggregator: NodeAggregatorConfig


class OdopObsConfig(BaseModel):
    process: ProcessProbeConfig
    system: SystemProbeConfig
    probe_connector: SocketConnectorConfig
    exporter: ExporterConfig



# Metrics are implemented based on these classes to be compatible with Prometheus
from .lang.datamodel_enum import MetricNameEnum


class PrometheusMetric:
    """
    This class defines the common attribute and provide basic function for handling a metric
    - Attribute:
        - name: name of the metric
        - description: describe the metric
        - value: value of the metric
        - category: group metric into specific category supporting building QoA_Report
        - others: (Developing)

    - Function:
        - set: set specific value
        - get_val: return current value
        - get_name: return metric name
        - get_des: return metric description
        - other: (Developing)

    - Category: metrics are categorized into following groups
        0 - Quality: Performance (metrics for evaluating service performance e.g., response time, throughput)
        1 - Quality: Data (metrics for evaluating data quality e.g., missing, duplicate, erroneous)
        2 - Quality: Inference (metrics for evaluating quality of ML inference, measured from inferences e.g., accuracy, confidence)
        3 - Resource: metrics for evaluating resource utilization e.g. CPU, Memory
        Other: To do (extend more categories)
    """

    def __init__(
        self, metric_name: MetricNameEnum, description, default_value=-1, category=0
    ):
        self.name = metric_name
        self.description = description
        self.default_value = default_value
        self.value = default_value
        self.category = category

    def set(self, value):
        self.value = value

    def get_val(self):
        return self.value

    def get_name(self):
        return self.name

    def get_des(self):
        return self.description

    def get_category(self):
        return self.category

    def reset(self):
        self.value = self.default_value

    def __str__(self) -> str:
        return "metric_name: " + self.name + ", " + "value: " + str(self.value)

    def to_dict(self):
        mectric_dict = {}
        mectric_dict[self.name] = self.value
        return mectric_dict


class Counter(PrometheusMetric):
    """
    This class inherit all attributes of Metric
    - Attribute: (Developing)

    - Function:
        - inc: increase its value by num
        - reset: set the value back to zero
        - others: (Developing)
    """

    def __init__(self, metric_name, description, default_value=0, category=0):
        super().__init__(metric_name, description, default_value, category)

    def inc(self, num=1):
        self.value += num


class Gauge(PrometheusMetric):
    """
    This class inherit all attributes of Metric
    - Attribute: (Developing)

    - Function:
        - inc: increase its value by num
        - others: (Developing)
    """

    def __init__(self, metric_name, description, default_value=-1, category=0):
        super().__init__(metric_name, description, default_value, category)

    def inc(self, num=1):
        self.value += num

    # TO DO:
    # implement other functions
    def dec(self, num=1):
        self.value -= num


class Summary(PrometheusMetric):
    """
    This class inherit all attributes of Metric
    - Attribute: (Developing)

    - Function:
        - inc: increase its value by num
        - others: (Developing)
    """

    def __init__(self, metric_name, description, default_value=-1, category=0):
        super().__init__(metric_name, description, default_value, category)

    def inc(self, num):
        self.value += num

    # TO DO:
    # implement other functions
    def dec(self, num):
        self.value -= num


class Histogram(PrometheusMetric):
    """
    This class inherit all attributes of Metric
    - Attribute: (Developing)

    - Function:
        - inc: increase its value by num
        - others: (Developing)
    """

    def __init__(self, metric_name, description, default_value=-1, category=0):
        super().__init__(metric_name, description, default_value, category)

    def inc(self, num):
        self.value += num

    # TO DO:
    # implement other functions
    def dec(self, num):
        self.value -= num



# NOTE: Mark lazy module globally, only load when used
from __future__ import annotations

import lazy_import

lazy_import.lazy_module("requests")
lazy_import.lazy_module("fastapi")
lazy_import.lazy_module("tinyflux")
lazy_import.lazy_module("psutil")



# This library is built based on ydata_quality: https://github.com/ydataai/ydata-quality

# import io
# import pathlib
# import sys
# import traceback
#
# import numpy as np
# import pandas as pd
#
# from ..lang.datamodel_enum import ImageQualityNameEnum
# from ..utils.logger import qoa_logger
# from ..utils.qoa_utils import is_numpyarray
#
# p_dir = pathlib.Path(__file__).parent.parent.absolute()
# sys.path.append(str(p_dir))
#
# # Define metric names, return formats: dictionary {metric name} {sub-element}
# # Return error/debugging
# ################################################ DATA QUALITY ########################################################
#
#
# def eva_erronous(data, errors=None):
#     """
#     Return number/percentage of error data
#     data: numpy array or pandas data frame
#     errors: list of item considered as error
#     ratio: return percentage if set to True
#     sum: sum the result if set to True, otherwise return errors following the categories in list of 'errors'
#     """
#     try:
#         if "ErroneousDataIdentifier" not in globals():
#             global ErroneousDataIdentifier
#             from ydata_quality.erroneous_data import ErroneousDataIdentifier
#         if is_numpyarray(data):
#             data = pd.DataFrame(data)
#         if is_pddataframe(data):
#             if errors and isinstance(errors, list):
#                 eva_err = ErroneousDataIdentifier(df=data, ed_extensions=errors)
#             else:
#                 eva_err = ErroneousDataIdentifier(df=data)
#             error_df = eva_err.predefined_erroneous_data()
#
#             total_count = data.count().to_numpy().flatten().sum()
#             results = {}
#             results["totalErrors"] = error_df.to_numpy().flatten().sum()
#             results["errorRatio"] = 100 * error_df / total_count
#             return results
#         else:
#             qoa_logger.warning(f"Unsupported data: {type(data)}")
#             return None
#     except Exception as e:
#         qoa_logger.error(f"Error {type(e)} in eva_erronous: {e.__traceback__}")
#         traceback.print_exception(*sys.exc_info())
#
#
# def eva_duplicate(data):
#     """
#     Return data/percentage of duplicate
#     data: numpy array or pandas data frame
#     ratio: return percentage if set to True
#     """
#     try:
#         if "DuplicateChecker" not in globals():
#             global DuplicateChecker
#             from ydata_quality.duplicates import DuplicateChecker
#         if is_numpyarray(data):
#             data = pd.DataFrame(data)
#         if is_pddataframe(data):
#             dc = DuplicateChecker(df=data)
#             dc_eva = dc.exact_duplicates()
#             results = {}
#             results["duplicateRatio"] = 100 * len(dc_eva.index) / len(data.index)
#             results["totalDuplicate"] = len(dc_eva.index)
#             return results
#         else:
#             qoa_logger.warning(f"Unsupported data: {type(data)}")
#             return None
#     except Exception as e:
#         qoa_logger.error(f"Error {type(e)} in eva_duplicate: {e.__traceback__}")
#         traceback.print_exception(*sys.exc_info())
#
#
# def eva_missing(
#     data, null_count=True, correlations=False, predict=False, random_state=0
# ):
#     try:
#         if "MissingsProfiler" not in globals():
#             global MissingsProfiler
#             from ydata_quality.missing import MissingsProfiler
#         if is_numpyarray(data):
#             data = pd.DataFrame(data)
#         if is_pddataframe(data):
#             mp = MissingsProfiler(df=data, random_state=random_state)
#             results = {}
#             if null_count:
#                 results["nullCount"] = mp.null_count()
#             if correlations:
#                 results["correlations"] = mp.missing_correlations()
#             if predict:
#                 results["missingPrediction"] = mp.predict_missings()
#             return results
#         else:
#             qoa_logger.warning(f"Unsupported data: {type(data)}")
#             return None
#     except Exception as e:
#         qoa_logger.error(f"Error {type(e)} in eva_erronous: {e.__traceback__}")
#         traceback.print_exception(*sys.exc_info())
#
#
# class OutlierDetector:
#     def __init__(self, data):
#         self.data = None
#         self.update_data(data)
#
#     def detect_outlier(self, n_data, labels=None, random_state=0, n=10, cluster=False):
#         if labels is None:
#             labels = []
#         if is_numpyarray(n_data):
#             n_data = pd.DataFrame(n_data)
#         if is_pddataframe(n_data):
#             if self.data is not None:
#                 data = None
#                 try:
#                     data = pd.concat([self.data, n_data])
#                 except Exception as e:
#                     qoa_logger.error(
#                         f"Error {type(e)} in concatenating data: {e.__traceback__}"
#                     )
#                     traceback.print_exception(*sys.exc_info())
#                 if data is not None:
#                     results = {}
#                     for label in labels:
#                         try:
#                             if "LabelInspector" not in globals():
#                                 global LabelInspector
#                                 from ydata_quality.labelling import LabelInspector
#                             li = LabelInspector(
#                                 df=data, label=label, random_state=random_state
#                             )
#                             results[label] = li.outlier_detection(
#                                 th=n, use_clusters=cluster
#                             )
#                         except Exception as e:
#                             qoa_logger.error(
#                                 f"Error {type(e)} in LabelInspector: {e.__traceback__}"
#                             )
#                             traceback.print_exception(*sys.exc_info())
#                     return results
#                 else:
#                     return {"Error": "Cannot concatenate data"}
#             else:
#                 return {"Error": "Historical data has not been set"}
#         else:
#             return {"Error": f"Unsupported data: {type(data)}"}
#
#     def update_data(self, data):
#         if is_numpyarray(data):
#             data = pd.DataFrame(data)
#         if is_pddataframe(data):
#             self.data = data
#             return {"Response": "Success"}
#         else:
#             return {"Error": f"Unsupported data: {type(data)}"}
#
#
# def image_quality(image):
#     if "PIL" not in globals():
#         global PIL
#         import PIL
#     quality = {}
#     if isinstance(image, bytes):
#         image = PIL.Image.open(io.BytesIO(image))
#     if isinstance(image, np.ndarray):
#         image = PIL.Image.fromarray(image)
#     if isinstance(image, (PIL.Image.Image, PIL.JpegImagePlugin.JpegImageFile)):
#         # qoa_logger.debug(dir(image)
#         quality[ImageQualityNameEnum.image_size] = image.size
#         quality[ImageQualityNameEnum.color_mode] = image.mode
#         quality[ImageQualityNameEnum.color_channel] = len(image.getbands())
#     return quality
#
#
# def eva_none(data):
#     try:
#         if is_pddataframe(data):
#             data = data.to_numpy()
#         if is_numpyarray(data):
#             valid_count = np.count_nonzero(~np.isnan(data))
#             none_count = np.count_nonzero(np.isnan(data))
#             results = {}
#             results["totalValid"] = valid_count
#             results["totalNone"] = none_count
#             results["noneRatio"] = valid_count / (valid_count + none_count)
#             return results
#         else:
#             qoa_logger.warning(f"Unsupported data: {type(data)}")
#             return None
#     except Exception as e:
#         qoa_logger.error(f"Error {type(e)} in eva_none: {e.__traceback__}")
#         traceback.print_exception(*sys.exc_info())



import glob
import json
import logging
import os
import pathlib
import re
import shlex
import subprocess
import sys
import time
import traceback
from threading import Thread

import numpy as np
import psutil
import yaml

from .logger import qoa_logger


def make_folder(temp_path):
    try:
        if os.path.exists(temp_path):
            pass
        else:
            os.makedirs(temp_path)
        return True
    except Exception:
        return False


def get_cgroup_version() -> str:
    proc1 = subprocess.Popen("mount", stdout=subprocess.PIPE)
    proc2 = subprocess.Popen(
        shlex.split("grep cgroup"),
        stdin=proc1.stdout,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    if proc1.stdout:
        proc1.stdout.close()
    out, _ = proc2.communicate()
    if "cgroup2" in out.decode():
        return "v2"
    return "v1"


if get_cgroup_version() == "v2":
    CGROUP_VERSION = "v2"
else:
    CGROUP_VERSION = "v1"


def set_logger_level(logging_level):
    if logging_level == 0:
        log_level = logging.NOTSET
    elif logging_level == 1:
        log_level = logging.DEBUG
    elif logging_level == 2:
        log_level = logging.INFO
    elif logging_level == 3:
        log_level = logging.WARNING
    elif logging_level == 4:
        log_level = logging.ERROR
    elif logging_level == 5:
        log_level = logging.CRITICAL
    else:
        raise ValueError(f"Error logging level {logging_level}")
    qoa_logger.setLevel(log_level)


def load_config(file_path: str) -> dict:
    """
    file_path: file path to load config

    """
    try:
        if "json" in file_path:
            with open(file_path) as f:
                return json.load(f)
        if ("yaml" in file_path) or ("yml" in file_path):
            with open(file_path) as f:
                return yaml.safe_load(f)
        else:
            qoa_logger.warning("Unsupported format")
            return None
    except Exception:
        qoa_logger.error("Unable to load configuration")

    return None


def to_json(file_path: str, conf: dict):
    """
    file_path: file path to save config
    """
    with open(file_path, "w") as f:
        json.dump(conf, f)


def to_yaml(file_path: str, conf: dict):
    """
    file_path: file path to save config
    """
    with open(file_path, "w") as f:
        yaml.dump(conf, f)


def get_sys_cpu():
    stats = psutil.cpu_stats()
    cpu_time = psutil.cpu_times()
    info = {}
    for key in stats._fields:
        info[key] = getattr(stats, key)
    for key in cpu_time._fields:
        info[key] = getattr(cpu_time, key)
    return info


def get_sys_cpu_util():
    info = {}
    core_utils = psutil.cpu_percent(percpu=True)
    for core_num, core_util in enumerate(core_utils):
        info[f"core_{core_num}"] = core_util
    return info


def get_sys_cpu_metadata():
    cpu_freq = psutil.cpu_freq()
    frequency = {"value": cpu_freq.max / 1000, "unit": "GHz"}
    cpu_threads = psutil.cpu_count(logical=True)
    return {"frequency": frequency, "thread": cpu_threads}


def get_sys_mem():
    stats = psutil.virtual_memory()
    info = {}
    for key in stats._fields:
        info[key] = getattr(stats, key)
    return info


def get_sys_net():
    info = {}
    net = psutil.net_io_counters()
    for key in net._fields:
        info[key] = getattr(net, key)
    return info


def report_proc_cpu(process):
    report = {}
    cpu_time = process.cpu_times()
    context = process.num_ctx_switches()
    for key in cpu_time._fields:
        report[key] = getattr(cpu_time, key)
    for key in context._fields:
        report[key] = getattr(context, key)
    report["num_thread"] = process.num_threads()

    return report


def report_proc_child_cpu(process: psutil.Process):
    # WARNING: this children function takes a lot of time
    child_processes = process.children(recursive=True)
    child_processes_count = len(child_processes)
    child_processes_cpu = {}
    process_cpu_time = process.cpu_times()
    for id, child_proc in enumerate(child_processes):
        cpu_time = child_proc.cpu_times()
        child_processes_cpu[f"child_{id}"] = float(cpu_time.user + cpu_time.system)

    total_cpu_usage = sum(child_processes_cpu.values())
    total_cpu_usage += float(
        process_cpu_time.user
        + process_cpu_time.system
        + process_cpu_time.children_user
        + process_cpu_time.children_system
    )

    return {
        "child_process": child_processes_count,
        "value": child_processes_cpu,
        "total": total_cpu_usage,
        "unit": "cputime",
    }


def get_proc_cpu(pid=None):
    if pid is None:
        pid = os.getpid()
    process = psutil.Process(pid)
    child_list = process.children()
    info = {}
    info[pid] = report_proc_cpu(process)

    for child in child_list:
        info[child.pid + "c"] = report_proc_cpu(child)
    return info


def report_proc_mem(process: psutil.Process):
    report = {}
    mem_info = process.memory_info()
    for key in mem_info._fields:
        report[key] = getattr(mem_info, key)
    return report


def get_proc_mem(pid=None):
    if pid is None:
        pid = os.getpid()
    process = psutil.Process(pid)
    child_list = process.children()
    info = {}
    info[pid] = report_proc_mem(process)

    for child in child_list:
        info[child.pid + "c"] = report_proc_mem(child)
    return info


def convert_to_gbyte(value):
    return value / 1024.0 / 1024.0 / 1024.0


def convert_to_mbyte(value):
    return value / 1024.0 / 1024.0


def convert_to_kbyte(value):
    return value / 1024.0


###################### SYSTEM REPORT ######################

sys_monitor_flag = False
process_monitor_flag = False
doc_monitor_flag = False


def system_report(client, interval: int, to_mb=True, to_gb=False, to_kb=False):
    report = {}
    last_net_value = {"sent": 0, "receive": 0}
    while sys_monitor_flag:
        try:
            report["sys_cpu_stats"] = get_sys_mem()
        except Exception as e:
            qoa_logger.error(f"Error {type(e)} in report CPU stat: {e.__traceback__}")
            traceback.print_exception(*sys.exc_info())
        try:
            report["sys_mem_stats"] = get_sys_mem()
        except Exception as e:
            qoa_logger.error(
                f"Error {type(e)} in report memory stat: {e.__traceback__}"
            )
            traceback.print_exception(*sys.exc_info())
        try:
            report["sys_net_stats"] = get_sys_net()
            sent = 0
            if to_mb:
                sent = convert_to_mbyte(psutil.net_io_counters().bytes_sent)
                receive = convert_to_mbyte(psutil.net_io_counters().bytes_recv)
            elif to_gb:
                sent = convert_to_gbyte(psutil.net_io_counters().bytes_sent)
                receive = convert_to_gbyte(psutil.net_io_counters().bytes_recv)
            elif to_kb:
                sent = convert_to_kbyte(psutil.net_io_counters().bytes_sent)
                receive = convert_to_kbyte(psutil.net_io_counters().bytes_recv)
            else:
                sent = psutil.net_io_counters().bytes_sent
                receive = psutil.net_io_counters().bytes_recv
            curr_net_value = {"sent": sent, "receive": receive}
            report["sys_net_send"] = curr_net_value["sent"] - last_net_value["sent"]
            report["sys_net_receive"] = (
                curr_net_value["receive"] - last_net_value["receive"]
            )
            last_net_value = curr_net_value.copy()
        except Exception as e:
            qoa_logger.error(
                f"Error {type(e)} in report network stat: {e.__traceback__}"
            )
            traceback.print_exception(*sys.exc_info())
        try:
            client.report(report=report)
        except Exception as e:
            qoa_logger.error(
                f"Error {type(e)} in sent system report: {e.__traceback__}"
            )
            traceback.print_exception(*sys.exc_info())
        time.sleep(interval)


def sys_monitor(client, interval: int):
    sub_thread = Thread(target=system_report, args=(client, interval))
    sub_thread.start()


###################### PROCESS REPORT ######################

# def process_report(client, interval:int, pid:int = None):
#     report = {}
#     while procMonitorFlag:
#         try:
#             report["proc_cpu_stats"] = get_proc_cpu()
#         except Exception as e:
#             qoaLogger.error("Error {} in report process cpu stat: {}".format(type(e),e.__traceback__))
#             traceback.print_exception(*sys.exc_info())
#         try:
#             report["proc_mem_stats"] = get_proc_mem()
#         except Exception as e:
#             qoaLogger.error("Error {} in report process memory stat: {}".format(type(e),e.__traceback__))
#             traceback.print_exception(*sys.exc_info())
#         try:
#             client.report(report=report)
#         except Exception as e:
#             qoaLogger.error("Error {} in sent process report: {}".format(type(e),e.__traceback__))
#             traceback.print_exception(*sys.exc_info())
#         time.sleep(interval)


# def process_monitor(client, interval:int, pid:int = None):
#     if (pid == None):
#         pid = os.getpid()
#     sub_thread = Thread(target=process_report, args=(client, interval, pid))
#     sub_thread.start()


###################### DOCKER REPORT ######################


def get_cpu_stat(stats, key):
    if key == "percentage":
        usage_delta = (
            stats["cpu_stats"]["cpu_usage"]["total_usage"]
            - stats["precpu_stats"]["cpu_usage"]["total_usage"]
        )

        system_delta = (
            stats["cpu_stats"]["system_cpu_usage"]
            - stats["precpu_stats"]["system_cpu_usage"]
        )

        len_cpu = stats["cpu_stats"]["online_cpus"]

        percentage = (usage_delta / system_delta) * len_cpu * 100
        return round(percentage, 2)
    else:
        return -1


def get_mem_stat(stats, key):
    if key == "used":
        return stats["memory_stats"]["usage"]
    else:
        return -1


def merge_report(f_report, i_report, prio=True):
    try:
        if isinstance(f_report, dict) and isinstance(i_report, dict):
            key_list = tuple(f_report.keys())
            for key in key_list:
                if key in i_report:
                    f_report[key] = merge_report(f_report[key], i_report[key], prio)
                    i_report.pop(key)
            f_report.update(i_report)
        elif f_report != i_report:
            if prio is True:
                return f_report
            else:
                return i_report
    except Exception as e:
        qoa_logger.error(f"Error {type(e)} in mergeReport: {e.__traceback__}")
        traceback.print_exception(*sys.exc_info())
    return f_report


def get_dict_at(dict, i=0):
    try:
        keys = list(dict.keys())
        return keys[i], dict[keys[i]]
    except Exception as e:
        qoa_logger.error(f"Error {type(e)} in get_dict_at: {e.__traceback__}")
        traceback.print_exception(*sys.exc_info())


def get_file_dir(file, to_string=True):
    current_dir = pathlib.Path(file).parent.absolute()
    if to_string:
        return str(current_dir)
    else:
        return current_dir


def get_parent_dir(file, parent_level=1, to_string=True):
    current_dir = get_file_dir(file=file, to_string=False)
    for _ in range(parent_level):
        current_dir = current_dir.parent.absolute()
    if to_string:
        return str(current_dir)
    else:
        return current_dir


def is_numpyarray(obj):
    return isinstance(obj, np.ndarray)


def get_process_allowed_cpus():
    # NOTE: 0 as PID represents the calling process
    pid = 0
    affinity = os.sched_getaffinity(pid)
    return list(affinity)


def get_process_allowed_memory():
    if CGROUP_VERSION == "v1":
        with open("/proc/self/cgroup") as file:
            for line in file:
                parts = line.strip().split(":")
                if len(parts) == 3 and parts[1] == "memory":
                    cgroup_path = parts[2]
                    memory_limit_file = re.sub(r"/task_\d+", "", cgroup_path)

                    number_of_task = len(
                        glob.glob(f"/sys/fs/cgroup/memory{memory_limit_file}/task_*")
                    )

                    with open(
                        f"/sys/fs/cgroup/memory{memory_limit_file}/memory.limit_in_bytes"
                    ) as limit_file:
                        memory_limit_str = limit_file.read().strip()
                        try:
                            memory_limit_int = int(memory_limit_str)
                            return memory_limit_int / number_of_task
                        except ValueError:
                            return memory_limit_str
            return None

    else:
        with open("/proc/self/cgroup") as file:
            for line in file:
                parts = line.strip().split(":")
                cgroup_path = parts[2]
                pattern = r"/task_\d+"
                cgroup_path = re.sub(pattern, "", cgroup_path)
                with open(f"/sys/fs/cgroup{cgroup_path}/memory.max") as limit_file:
                    number_of_task = len(
                        glob.glob(f"/sys/fs/cgroup{cgroup_path}/task_*")
                    )
                    memory_limit_str = limit_file.read().strip()
                    try:
                        memory_limit_int = int(memory_limit_str)
                        return memory_limit_int / number_of_task
                    except ValueError:
                        return memory_limit_str
            return None



import json
import time

import docker

from qoa4ml.reports.resources_report_model import DockerReport

from ..config.configs import ClientInfo, DockerProbeConfig
from ..connector.base_connector import BaseConnector
from ..utils.docker_util import get_docker_stats
from ..utils.logger import qoa_logger
from .probe import Probe


class DockerMonitoringProbe(Probe):
    def __init__(
        self,
        config: DockerProbeConfig,
        connector: BaseConnector,
        client_info: ClientInfo,
    ) -> None:
        super().__init__(config, connector, client_info)
        self.config = config
        if self.config.require_register:
            self.obs_service_url = self.config.obs_service_url
        self.docker_client = docker.from_env()

    def create_report(self):
        try:
            reports = get_docker_stats(self.docker_client, self.config.container_list)
            docker_report = DockerReport(
                metadata=self.client_info,
                timestamp=time.time(),
                container_reports=reports,
            )
            reports_dict = docker_report.model_dump()
            # NOTE: if the reports dict is empty, the loop will run very fast, so here add 2 seconds as if there is container to report
            if not reports_dict:
                time.sleep(2)
            return json.dumps(reports_dict)
        except RuntimeError:
            qoa_logger.exception(
                "Maybe running in the background result in this exception!"
            )
        return json.dumps({"error": "RuntimeError"})



import logging

import uvicorn
from fastapi import FastAPI

from ...config.configs import ExporterConfig
from .node_aggregator import NodeAggregator

logging.basicConfig(
    format="%(asctime)s:%(levelname)s -- %(message)s", level=logging.INFO
)


class Exporter:
    def __init__(self, config: ExporterConfig) -> None:
        self.app = FastAPI()
        self.config = config
        self.node_aggregator = NodeAggregator(self.config.node_aggregator)
        self.app.include_router(self.node_aggregator.router)

    def start(self):
        self.node_aggregator.start()
        uvicorn.run(self.app, host=self.config.host, port=self.config.port)



import logging
import pickle
import socket
import time
from typing import Optional

from ..config.configs import SocketConnectorConfig
from .base_connector import BaseConnector


class SocketConnector(BaseConnector):
    def __init__(self, config: SocketConnectorConfig):
        self.config = config
        self.host = config.host
        self.port = config.port

    def send_report(self, body_message: str, log_path: Optional[str] = None):
        try:
            start = time.time()
            client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            client_socket.connect((self.host, self.port))
            serialized_message = pickle.dumps(body_message)
            client_socket.sendall(serialized_message)
            client_socket.close()
            if log_path:
                with open(log_path, "a", encoding="utf-8") as file:
                    file.write(str((time.time() - start) * 1000) + "\n")
        except ConnectionRefusedError:
            logging.error("Connection to aggregator refused")



import time
from threading import Event, Thread


class RepeatedTimer:
    """Repeat `function` every `interval` seconds."""

    def __init__(self, interval, function, *args, **kwargs):
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.start = time.time()
        self.event = Event()
        self.thread = Thread(target=self._target, daemon=True)
        self.thread.start()

    def _target(self):
        self.function(*self.args, **self.kwargs)
        while not self.event.wait(self._time):
            self.function(*self.args, **self.kwargs)

    @property
    def _time(self):
        return self.interval - ((time.time() - self.start) % self.interval)

    def stop(self):
        self.event.set()
        self.thread.join()
